from ribosome.machine import may_handle
from ribosome.unite import UniteKind
from ribosome.unite.machine import UniteMachine, UniteTransitions

from amino import Map, List
from amino.lazy import lazy

from myo.plugins.unite.message import UniteHistory, UniteCommands
from myo.state import MyoTransitions, MyoComponent
from myo.plugins.unite.data import UniteNames, HistorySource, CommandsSource


class MyoUniteTransitions(UniteTransitions, MyoTransitions):

    def unite_cmd(self, cmd):
        args = ' '.join(self.msg.unite_args)
        self.vim.cmd('Unite {} {}'.format(cmd, args))

    @may_handle(UniteCommands)
    def commands(self):
        self.unite_cmd(UniteNames.commands)

    @may_handle(UniteHistory)
    def history(self):
        self.unite_cmd(UniteNames.history)


class Plugin(UniteMachine, MyoComponent):
    Transitions = MyoUniteTransitions

    def __init__(self, *a, **kw):
        UniteMachine.__init__(self)
        MyoComponent.__init__(self, *a, **kw)

    @lazy
    def history_source(self):
        return HistorySource()

    @lazy
    def commands_source(self):
        return CommandsSource()

    @property
    def sources(self):
        return List(self.history_source, self.commands_source)

    @lazy
    def run_action(self):
        return Map(name='run', handler=UniteNames.run, desc='run command')

    @property
    def actions(self):
        return List(self.run_action)

    @lazy
    def run_kind(self):
        return UniteKind(UniteNames.command, List(self.run_action))

    @property
    def kinds(self):
        return List(self.run_kind)

__all__ = ('Plugin',)



from functools import singledispatch  # type: ignore

from ribosome.machine import Error

from myo.dispatch import Dispatcher
from myo.command import ShellCommand, Shell, CommandJob
from myo.plugins.tmux.message import (TmuxRunCommand, TmuxRunShell,
                                      TmuxRunTransient)


@singledispatch
def convert_message(msg, options):
    return Error('invalid message for tmux dispatch: {}'.format(msg))


@convert_message.register(Shell)
def _convert_shell(cmd, options):
    return TmuxRunShell(CommandJob(command=cmd), options)


@convert_message.register(ShellCommand)
def _convert_shell_command(cmd, options):
    return TmuxRunCommand(CommandJob(command=cmd), options)


@convert_message.register(CommandJob)
def _convert_command_job(cmd, options):
    return TmuxRunCommand(cmd, options)


class TmuxDispatcher(Dispatcher):

    def can_run(self, cmd):
        return isinstance(cmd, (ShellCommand, CommandJob))

    def message(self, cmd, options):
        return convert_message(cmd, options)

__all__ = ('TmuxDispatcher',)



from amino import List, __, Just, _
from amino.test import later
from amino.lazy import lazy

from ribosome.test.unite import unite
from ribosome.record import encode_json

from myo.command import ShellCommand, CommandJob, TransientCommandJob
from myo.plugins.unite.format import unite_format, unite_format_str_command
from myo.plugins.command.main import CommandTransitions

from integration._support.command import CmdSpec

indent = ' {}'.format


class _UniteSpecBase(CmdSpec):

    @lazy
    def _name1(self):
        return 'com1'

    @lazy
    def _name2(self):
        return 'com2'

    @lazy
    def _cmd(self):
        return 'ls'

    def _create_commands(self):
        self._create_command(self._name1, self._cmd)
        self._create_command(self._name2, self._cmd)

    @property
    def _plugins(self):
        return super()._plugins + List('myo.plugins.unite',
                                       'integration._support.plugins.dummy')

    def _format(self, name, line):
        return indent(unite_format_str_command.format(name=name, line=line))


class UniteLoadHistorySpec(_UniteSpecBase):

    @property
    def _cmd(self):
        return 'test1'

    @lazy
    def _history(self):
        cmd = ShellCommand(name='other', line='ls')
        return List(
            CommandJob(command=ShellCommand(name=self._cmd, line='tee')),
            TransientCommandJob(command=cmd, override_line='tail')
        )

    def _set_vars(self):
        super()._set_vars()
        self.vim.vars.set('Myo_history',
                          encode_json(self._history).get_or_raise)

    @property
    def _last(self):
        return (lambda: self.vim.vars.pd('last_command') // __.get('name'))

    @unite
    def history(self):
        self._create_command(self._cmd, 'ls')
        self.vim.cmd_sync('MyoUniteHistory')
        target = self._history / unite_format / _['word'] / indent
        later(lambda: self.vim.buffer.content.should.equal(target))


class UniteHistorySpec(_UniteSpecBase):

    @unite
    def history(self):
        self._create_commands()
        self._run_command(self._name1)
        self._run_command(self._name2)
        later(lambda: self._last().should.contain(self._name2))
        self.vim.cmd_sync('MyoUniteHistory')
        target = List(self._format(self._name2, self._cmd),
                      self._format(self._name1, self._cmd))
        later(lambda: self.vim.buffer.content.should.equal(target))
        self.vim.window.set_cursor(2)
        self.vim.feedkeys('\r')
        later(lambda: self._last().should.contain(self._name2))
        self.vim.cmd_sync('MyoUniteHistory')
        later(lambda: self.vim.buffer.content.should.equal(target.reversed))

_test_line = 'echo \'testing\''


def _test_ctor():
    return Just(_test_line)


class UniteTestSpec(_UniteSpecBase):

    @unite
    def unite_test_run(self):
        self.json_cmd_sync('MyoTest',
                           ctor='py:integration.unite_spec._test_ctor')
        later(lambda: (self._last() | '').should.contain('test_line'))
        self.vim.vars.set_p('last_command', {})
        self.vim.cmd_sync('MyoUniteHistory')
        target = unite_format_str_command.format(name='<test>',
                                                 line=_test_line)
        self._buffer_content(List(indent(target)))
        self.vim.feedkeys('\r')
        later(lambda: (self._last() | '').should.contain('test_line'))


class UniteCommandsSpec(_UniteSpecBase):

    def _pre_start(self):
        super()._pre_start()
        self.vim.vars.set_p(CommandTransitions._test_cmd_var, False)

    @unite
    def commands(self):
        self._create_commands()
        self.vim.cmd_sync('MyoUniteCommands')
        target = List(self._format(name=self._name1, line=self._cmd),
                      self._format(name=self._name2, line=self._cmd))
        later(lambda: self.vim.buffer.content.should.equal(target))
        self.vim.window.set_cursor(2)
        self.vim.feedkeys('\r')
        later(lambda: self._last().should.contain(self._name2))
        self.vim.cmd_sync('MyoUniteCommands')
        later(lambda: self.vim.buffer.content.should.equal(target))

__all__ = ('UniteLoadHistorySpec', 'UniteHistorySpec', 'UniteCommandsSpec')
